<?php

require('Unit.class.php');

class AggregateQuery {

    public $sql;

    protected $spec;

    function __construct( AggregateQuerySpec $spec ) {
        $this->spec = $spec;
        $this->build();
    }

    protected function getTable() {
        return $this->spec->table;
    }

    protected function build() {
        $this->sql = "select \n"
            . $this->buildDataSelects() ."\n"
            . $this->buildFrom() ."\n"
            . $this->buildWhere() ."\n"
            . $this->buildGroupBy() ."\n"
            . $this->buildOrderBy();
        $limit = $this->buildLimit();
        if( $limit ) $this->sql .= "\n$limit";
    }

    protected function buildDataSelects() {
        $selects = array();
        foreach( $this->spec->data as $data ) {
            // XXX do special handling for dateTime to be based off of the grouping instead?
            $selects[] = $this->buildDataSelect($data);
        }
        return join(",\n", $selects );
    }

    protected function buildDataSelect($data) {
        $field = $data->field;
        $agg = $data->agg;
        $unit = $data->unit;
        $decimals = $data->decimals;

        $select = $field;
        if( $this->isGrouped() ) {
            $select = "$agg($select)";
        }
        $select = $this->buildUnitConversion($select, $field, $unit);
        $alias = $this->buildDataSelectAlias($field, $agg);
        if($decimals !== NULL) {
            $select = "round($select, $decimals) $alias";
        }
        return $select;
    }

    protected function buildUnitConversion($value, $field, $unit) {
        if( $unit == NULL ) {
            return $value;
        }
        $columnUnit = $this->spec->entityConfig['columns'][$field]['unit'];
        return UnitConvert::getSqlFormula($value, $columnUnit, $unit);
    }

    protected function buildDataSelectAlias($baseAlias, $agg) {
        $alias = $baseAlias;
        if( $this->isGrouped() ) {
            $alias .= "_$agg";
        }
        return $alias;
    }

    protected function buildFrom() {
        return "from {$this->getTable()}";
    }

    protected function buildWhere() {
        $where = array();
        $start = $this->spec->start;
        $end = $this->spec->end;
        if( $start != NULL ) {
            $where[] = "{$this->spec->dateTimeColumn} >= $start";
        }
        if( $end != NULL ) {
            $where[] = "{$this->spec->dateTimeColumn} <= $end";
        }
        return 
            count($where) > 0 ? 
            "where " . join(" and ", $where) . "\n" :
            "";
    }

    protected function buildGroupBy() {
        if( $this->isGrouped() && !$this->isSingleGroup() ) {
            return "group by " . $this->buildGroup() . "\n";
        }
        return "";
    }

    protected function buildGroup() {
        if( $this->isGrouped() ) {
            $slice = $this->buildGroupSlice();
            // FIXME cast: sqlite uses int, mysql uses unsigned
            return "cast(({$this->spec->dateTimeColumn} / $slice) as unsigned)";
        }
        return NULL;
    }

    protected function buildGroupSlice() {
        $value = $this->spec->group->value;
        // groups
        if( $this->spec->group->type == GroupType::groups ) {
            $start = $this->spec->start;
            $end = $this->spec->end;
            // the sub-select for the min/max is an SQLite query optimization
            // http://www.sqlite.org/optoverview.html#minmax
            $start = ($start == NULL) ? $this->buildSubSelect( "min({$this->spec->dateTimeColumn})" ) : $start;
            $end = ($end == NULL) ? $this->buildSubSelect( "max({$this->spec->dateTimeColumn})" ) : $end;
            // i.e. cast((dateTime / (select (max(dateTime) - min(dateTime)) / 5.0 from archive) ) as int)
            //$slice = "(select ($end - $start) / cast($value as real) from {$this->getTable()})";
            // FIXME cast: sqlite uses REAL, mysql uses decimal
            $slice = "(($end - $start) / cast($value as decimal))";
        }
        // period
        else {
            $slice = $value;
        }
        return $slice;
    }

    protected function buildSubSelect($expression) {
        return "(select $expression from {$this->getTable()})";
    }

    protected function buildOrderBy() {
        // always ordered by dateTime for now
        return "order by {$this->spec->dateTimeColumn} {$this->spec->order}";
    }

    protected function buildLimit() {
        $limit = $this->spec->limit;
        return $limit ? "limit $limit" : NULL;
    }

    protected function isGrouped() {
        return $this->spec->isGrouped();
    }

    protected function isSingleGroup() {
        return $this->spec->group->isSingle();
    }
}

?>
